import numpy as np

from nabirds import utils

from .base import BasePartCollection, BasePart
from .annotation import BBoxPart

class Parts(BasePartCollection):

	def __init__(self, image, part_annotations, rescale_size):
		super(Parts, self).__init__()
		if part_annotations is None:
			self._parts = []
		else:
			self._parts = [BasePart.new(image, a, rescale_size) for a in part_annotations]


class UniformParts(BasePartCollection):

	def __init__(self, image, ratio):
		super(UniformParts, self).__init__()
		self._parts = list(self.generate_parts(image, ratio))

	def generate_parts(self, im, ratio, round_op=np.floor):
		h, w, c = utils.dimensions(im)

		part_w = round_op(w * ratio).astype(np.int32)
		part_h = round_op(h * ratio).astype(np.int32)

		n, m = w // part_w, h // part_h

		# fit best possible part_w and part_h
		part_w = int(w / n)
		part_h = int(h / m)

		for i in range(n*m):
			row, col = np.unravel_index(i, (n, m))
			x, y = col * part_w, row * part_h

			yield BBoxPart(im, [i, x, y, part_w, part_h])



import numpy as np

from os.path import join

from nabirds.utils import _MetaInfo
from .base import BaseAnnotations


class CARS_Annotations(BaseAnnotations):
	name="CARS"

	@property
	def meta(self):
		info = _MetaInfo(
			images_folder="images",
			images_file="images.txt",
			labels_file="labels.txt",
			split_file="tr_ID.txt",
			bounding_boxes="bounding_boxes.txt",
			bounding_box_dtype=np.dtype([(v, np.int32) for v in "xywh"]),
		)

		info.structure = [
			[info.images_file, "_images"],
			[info.labels_file, "labels"],
			[info.split_file, "_split"],
			[info.bounding_boxes, "_bounding_boxes"],
		]
		return info

	def __init__(self, *args, **kwargs):
		super(CARS_Annotations, self).__init__(*args, **kwargs)
		# set labels from [1..N] to [0..N-1]
		self.labels -= 1

	def _load_split(self):
		assert self._split is not None, "Train-test split was not loaded!"
		uuid_to_split = {uuid: int(split) for uuid, split in zip(self.uuids, self._split)}
		self.train_split = np.array([uuid_to_split[uuid] for uuid in self.uuids], dtype=bool)
		self.test_split = np.logical_not(self.train_split)

	def _load_parts(self):
		self.part_names = {}
		self._load_bounding_boxes()

	def _load_bounding_boxes(self):
		assert self._bounding_boxes is not None, "Bouding boxes were not loaded!"

		uuid_to_bbox = {}
		for content in [i.split() for i in self._bounding_boxes]:
			uuid, bbox = content[0], content[1:]
			uuid_to_bbox[uuid] = [float(i) for i in bbox]

		self.bounding_boxes = np.array(
			[tuple(uuid_to_bbox[uuid]) for uuid in self.uuids],
			dtype=self.meta.bounding_box_dtype)

	def parts(self, *args, **kwargs):
		return None

	def bounding_box(self, uuid):
		return self.bounding_boxes[self.uuid_to_idx[uuid]].copy()



import numpy as np
from PIL.Image import Image as PIL_Image

def dimensions(im):
	if isinstance(im, np.ndarray):
		if im.ndim != 3:
			import pdb; pdb.set_trace()
		assert im.ndim == 3, "Only RGB images are currently supported!"
		return im.shape
	elif isinstance(im, PIL_Image):
		w, h = im.size
		c = len(im.getbands())
		# assert c == 3, "Only RGB images are currently supported!"
		return h, w, c
	else:
		raise ValueError("Unknown image instance ({})!".format(type(im)))

def asarray(im, dtype=np.uint8):
	if isinstance(im, np.ndarray):
		return im.astype(dtype)
	elif isinstance(im, PIL_Image):
		return np.asarray(im, dtype=dtype)
	else:
		raise ValueError("Unknown image instance ({})!".format(type(im)))



import numpy as np

from matplotlib import pyplot as plt
from abc import ABC, abstractproperty

from nabirds import utils

class BasePartCollection(ABC):

	def __getitem__(self, i):
		return self._parts[i]

	def __repr__(self):
		return repr(np.stack([p.as_annotation for p in self._parts]))

	@property
	def selected(self):
		return np.array([p.is_visible for p in self._parts], dtype=bool)

	@property
	def selected_idxs(self):
		return np.where(self.selected)[0]

	def select(self, idxs):
		if isinstance(idxs, np.ndarray) and idxs.dtype == bool:
			# a mask is present, so convert it to indeces
			idxs = np.where(idxs)[0]

		for p in self._parts:
			p.is_visible = p._id in idxs

	def hide_outside_bb(self, *bounding_box):
		for p in self._parts:
			p.hide_if_outside(*bounding_box)

	def invert_selection(self):
		self.select(np.logical_not(self.selected))

	def offset(self, dx, dy):
		for p in self._parts:
			p.x += dx
			p.y += dy

	def visible_locs(self):
		vis = [(p._id, p.xy) for p in self._parts if p.is_visible]
		idxs, xy = zip(*vis)
		return np.array(idxs), np.array(xy).T

	def visible_crops(self, *args, **kwargs):
		return np.array([p.crop(*args, **kwargs) for p in self._parts])

	def plot(self, cmap=plt.cm.jet, **kwargs):
		for i, p in enumerate(self._parts):
			p.plot(color=cmap(i/len(self._parts)), **kwargs)

	def reveal(self, im, ratio, *args, **kwargs):
		res = np.zeros_like(im)

		for part in self._parts:
			if not part.is_visible: continue
			x, y, crop = part.reveal(im, ratio=ratio, *args, **kwargs)
			h, w, _ = crop.shape
			res[y:y+h, x:x+w] = crop

		return res




class BasePart(ABC):

	def __repr__(self):
		return repr(self.as_annotation)

	@staticmethod
	def new(image, annotation, rescale_size=-1):
		from .annotation import BBoxPart, LocationPart
		if len(annotation) == 4:
			return LocationPart(image, annotation, rescale_size)
		elif len(annotation) == 5:
			return BBoxPart(image, annotation, rescale_size)
		else:
			raise ValueError("Unknown part annotation format: {}".format(annotation))

	def rescale(self, image, annotation, rescale_size):
		if rescale_size is not None and rescale_size > 0:
			h, w, c = utils.dimensions(image)
			scale = np.array([w, h]) / rescale_size
			xy = annotation[1:3]
			xy = xy * scale
			annotation[1:3] = xy

		return annotation

	@property
	def is_visible(self):
		return self._is_visible

	@is_visible.setter
	def is_visible(self, value):
		self._is_visible = bool(value)

	@property
	def xy(self):
		return np.array([self.x, self.y])

	def crop(self, im, w, h, padding_mode="edge", is_location=True):
		if not self.is_visible:
			_, _, c = utils.dimensions(im)
			return np.zeros((h, w, c), dtype=np.uint8)

		x, y = self.xy
		pad_h, pad_w = h // 2, w // 2

		padded_im = np.pad(im, [(pad_h, pad_h), (pad_w, pad_w), [0,0]], mode=padding_mode)
		x0, y0 = x + pad_w, y + pad_h

		if is_location:
			x0, y0 = x0 - w // 2, y0 - h // 2

		return padded_im[y0:y0+h, x0:x0+w]

	@abstractproperty
	def middle(self):
		raise NotImplementedError

	def plot(self, **kwargs):
		return

	def hide_if_outside(self, x, y, w, h):
		mid_x, mid_y = self.middle
		self.is_visible = ((x <= mid_x <= x+w) and (y <= mid_y <= y+h))



from os.path import join

from nabirds.utils import _MetaInfo
from .base import BaseAnnotations


class NAB_Annotations(BaseAnnotations):
	name="NABirds"

	@property
	def meta(self):
		info = _MetaInfo(
			images_folder="images",
			images_file="images.txt",
			labels_file="labels.txt",
			hierarchy_file="hierarchy.txt",
			split_file="train_test_split.txt",
			parts_file=join("parts", "part_locs.txt"),
			part_names_file=join("parts", "parts.txt"),
		)

		info.structure = [
			[info.images_file, "_images"],
			[info.labels_file, "labels"],
			[info.hierarchy_file, "hierarchy"],
			[info.split_file, "_split"],
			[info.parts_file, "_part_locs"],
			[info.part_names_file, "_part_names"],
		]
		return info



import numpy as np

try:
	from yaml import CLoader as Loader, CDumper as Dumper
except ImportError:
	from yaml import Loader, Dumper

import yaml

from os.path import isfile

def feature_file_name(subset, part_info, model_info):

	return "{subset}{suffix}.{model}.npz".format(
			subset=subset,
			suffix=part_info.feature_suffix,
			model=model_info.folder)

def read_info_file(fpath):
	with open(fpath) as f:
		return attr_dict(yaml.load(f, Loader=Loader))

def random_idxs(idxs, rnd=None, n_parts=None):

	if rnd is None or isinstance(rnd, int):
		rnd = np.random.RandomState(rnd)
	else:
		assert isinstance(rnd, np.random.RandomState), \
			"'rnd' should be either a random seed or a RandomState instance!"

	n_parts = n_parts or rnd.randint(1, len(idxs))
	res = rnd.choice(idxs, n_parts, replace=False)
	res.sort()
	return res

class attr_dict(dict):
	def __getattr__(self, name):
		if name in self:
			return self[name]
		else:
			return super(attr_dict, self).get(name)

	def __getitem__(self, key):
		res = super(attr_dict, self).__getitem__(key)

		if isinstance(res, dict):
			return attr_dict(res)
		return res

class _MetaInfo(object):
	def __init__(self, **kwargs):
		for name, value in kwargs.items():
			setattr(self, name, value)
		self.structure = []


from .image import asarray, dimensions



from .cub import CUB_Annotations
from .nab import NAB_Annotations
from .cars import CARS_Annotations

from cvargparse.utils import BaseChoiceType

class AnnotationType(BaseChoiceType):
	CUB200 = CUB_Annotations
	NAB = NAB_Annotations
	CARS = CARS_Annotations

	Default = CUB200



import numpy as np
import abc
import warnings
import logging
from os.path import join, isfile, isdir
from collections import defaultdict, OrderedDict

from nabirds.utils import read_info_file, feature_file_name
from nabirds.dataset import Dataset

class BaseAnnotations(abc.ABC):

	FEATURE_PHONY = dict(train=["train"], test=["test", "val"])

	def __init__(self, root_or_infofile, parts=None, feature_model=None):
		super(BaseAnnotations, self).__init__()
		self.part_type = parts
		self.feature_model = feature_model

		if isdir(root_or_infofile):
			self.info = None
			self.root = root_or_infofile
		elif isfile(root_or_infofile):
			self.root = self.root_from_infofile(root_or_infofile, parts)
		else:
			raise ValueError("Root folder or info file does not exist: \"{}\"".format(
				root_or_infofile
			))

		for fname, attr in self.meta.structure:
			self.read_content(fname, attr)

		self.labels = np.array([int(l) for l in self.labels], dtype=np.int32)

		self._load_uuids()
		self._load_parts()
		self._load_split()

	@property
	def data_root(self):
		if self.info is None: return None
		return join(self.info.BASE_DIR, self.info.DATA_DIR)

	@property
	def dataset_info(self):
		if self.info is None: return None
		if self.part_type is None:
			return self.info.DATASETS[self.__class__.name]
		else:
			return self.info.PARTS[self.part_type]

	def root_from_infofile(self, info_file, parts=None):
		self.info = read_info_file(info_file)

		dataset_info = self.dataset_info
		annot_dir = join(self.data_root, dataset_info.folder, dataset_info.annotations)

		assert isdir(annot_dir), "Annotation folder does exist! \"{}\"".format(annot_dir)
		return annot_dir

	def new_dataset(self, subset=None, dataset_cls=Dataset, **kwargs):
		if subset is not None:
			uuids = getattr(self, "{}_uuids".format(subset))
		else:
			uuids = self.uuids

		kwargs = self.check_parts_and_features(subset, **kwargs)
		return dataset_cls(uuids=uuids, annotations=self, **kwargs)

	def check_parts_and_features(self, subset, **kwargs):
		dataset_info = self.dataset_info
		if dataset_info is None:
			return kwargs
		logging.debug("Dataset info: {}".format(dataset_info))

		# TODO: pass all scales
		new_opts = {}

		if "scales" in dataset_info:
			new_opts["ratio"] = dataset_info.scales[0]

		if "is_uniform" in dataset_info:
			new_opts["uniform_parts"] = dataset_info.is_uniform

		if self.part_type is not None:
			new_opts["part_rescale_size"] = dataset_info.rescale_size

		if None not in [subset, self.feature_model]:
			tried = []
			model_info = self.info.MODELS[self.feature_model]
			for subset_phony in BaseAnnotations.FEATURE_PHONY[subset]:
				features = feature_file_name(subset_phony, dataset_info, model_info)
				feature_path = join(self.root, "features", features)
				if isfile(feature_path): break
				tried.append(feature_path)
			else:
				raise ValueError(
					"Could not find any features in \"{}\" for {} subset. Tried features: {}".format(
					join(self.root, "features"), subset, tried))
			new_opts["features"] = feature_path
		new_opts.update(kwargs)

		logging.debug("Final kwargs: {}".format(new_opts))
		return new_opts

	@property
	@abc.abstractmethod
	def meta(self):
		pass

	def _path(self, file):
		return join(self.root, file)

	def _open(self, file):
		return open(self._path(file))

	def read_content(self, file, attr):
		content = None
		fpath = self._path(file)
		if isfile(fpath):
			with self._open(file) as f:
				content = [line.strip() for line in f if line.strip()]
		else:
			warnings.warn("File \"{}\" was not found!".format(fpath))

		setattr(self, attr, content)


	def _load_uuids(self):
		assert self._images is not None, "Images were not loaded!"
		uuid_fnames = [i.split() for i in self._images]
		self.uuids, self.images = map(np.array, zip(*uuid_fnames))
		self.uuid_to_idx = {uuid: i for i, uuid in enumerate(self.uuids)}

	def _load_parts(self):
		assert self._part_locs is not None, "Part locations were not loaded!"
		# this part is quite slow... TODO: some runtime improvements?
		uuid_to_parts = defaultdict(list)
		for content in [i.split() for i in self._part_locs]:
			uuid_to_parts[content[0]].append([float(i) for i in content[1:]])

		self.part_locs = np.stack([uuid_to_parts[uuid] for uuid in self.uuids]).astype(int)

		if hasattr(self, "_part_names") and self._part_names is not None:
			self._load_part_names()

	def _load_part_names(self):
		self.part_names = OrderedDict()
		self.part_name_list = []
		for line in self._part_names:
			part_idx, _, name = line.partition(" ")
			self.part_names[int(part_idx)] = name
			self.part_name_list.append(name)

	def _load_split(self):
		assert self._split is not None, "Train-test split was not loaded!"
		uuid_to_split = {uuid: int(split) for uuid, split in [i.split() for i in self._split]}
		self.train_split = np.array([uuid_to_split[uuid] for uuid in self.uuids], dtype=bool)
		self.test_split = np.logical_not(self.train_split)

	def image_path(self, image):
		return join(self.root, self.meta.images_folder, image)

	def image(self, uuid):
		fname = self.images[self.uuid_to_idx[uuid]]
		return self.image_path(fname)

	def label(self, uuid):
		return self.labels[self.uuid_to_idx[uuid]].copy()

	def parts(self, uuid):
		return self.part_locs[self.uuid_to_idx[uuid]].copy()


	def _uuids(self, split):
		return self.uuids[split]

	@property
	def train_uuids(self):
		return self._uuids(self.train_split)

	@property
	def test_uuids(self):
		return self._uuids(self.test_split)




import numpy as np

from matplotlib.patches import Rectangle

from nabirds import utils
from .base import BasePart

class LocationPart(BasePart):
	DEFAULT_RATIO = np.sqrt(49 / 400) # 0.35

	def __init__(self, image, annotation, rescale_size=None):
		super(LocationPart, self).__init__()

		annotation = self.rescale(image, annotation, rescale_size)
		# here x,y are the center of the part
		self._id, self.x, self.y, self.is_visible = annotation
		self._ratio = LocationPart.DEFAULT_RATIO

	def as_annotation(self):
		return np.array([self._id, self.x, self.y, self.is_visible])

	def crop(self, image, ratio=None, padding_mode="edge", *args, **kwargs):
		ratio = ratio or self._ratio
		_h, _w, c = utils.dimensions(image)
		w, h = int(_w * ratio), int(_h * ratio)
		return super(LocationPart, self).crop(image, w, h,
				padding_mode, is_location=True)

	def reveal(self, im, ratio=None, *args, **kwargs):
		_h, _w, c = utils.dimensions(im)
		w, h = int(_w * ratio), int(_h * ratio)
		x,y = self.xy
		x, y = max(x - w // 2, 0), max(y - h // 2, 0)
		return x, y, im[y:y+h, x:x+w]

	def plot(self, im, ax, ratio, fill=False, linestyle="--", **kwargs):
		if not self.is_visible: return
		x, y = self.xy
		_h, _w, c = utils.dimensions(im)
		w, h = int(_w * ratio), int(_h * ratio)
		ax.add_patch(Rectangle(
			(x-w//2, y-h//2), w, h,
			fill=fill, linestyle=linestyle,
			**kwargs
		))

		ax.scatter(*self.middle, marker="x", color="white", alpha=0.8)


	@property
	def middle(self):
		return np.array([self.x, self.y])

class BBoxPart(BasePart):

	def __init__(self, image, annotation, rescale_size=None):
		super(BBoxPart, self).__init__()

		annotation = self.rescale(image, annotation, rescale_size)
		# here x,y are top left corner of the part
		self._id, self.x, self.y, self.w, self.h = annotation
		self.is_visible = True

	@property
	def as_annotation(self):
		return np.array([self._id, self.x, self.y, self.w, self.h])

	def rescale(self, image, annotation, rescale_size):
		if rescale_size is not None and rescale_size > 0:
			annotation = super(BBoxPart, self).rescale(image, annotation, rescale_size)
			h, w, c = utils.dimensions(image)
			scale = np.array([w, h]) / rescale_size
			wh = annotation[3:5]
			wh = wh * scale
			annotation[3:5] = wh

		return annotation

	@property
	def middle(self):
		return np.array([self.x + self.w // 2, self.y + self.h // 2])

	def crop(self, image, padding_mode="edge", *args, **kwargs):
		return super(BBoxPart, self).crop(image, self.w, self.h,
			padding_mode, is_location=False)

	def reveal(self, im, ratio, *args, **kwargs):
		_h, _w, c = utils.dimensions(im)
		x,y = self.xy
		return x, y, im[y:y+self.h, x:x+self.w]



	def plot(self, im, ax, ratio, fill=False, linestyle="--", **kwargs):
		ax.add_patch(Rectangle(
			(self.x, self.y), self.w, self.h,
			fill=fill, linestyle=linestyle,
			**kwargs
		))

		ax.scatter(*self.middle, marker="x", color="white", alpha=0.8)




from abc import ABC, abstractmethod

import numpy as np
import six

from matplotlib.patches import Rectangle

class BaseMixin(ABC):

	@abstractmethod
	def get_example(self, i):
		s = super(BaseMixin, self)
		if hasattr(s, "get_example"):
			return s.get_example(i)

	def plot_bounding_box(self, i, ax, fill=False, linestyle="--", **kwargs):
		x, y, w, h = self.bounding_box(i)
		ax.add_patch(Rectangle(
			(x,y), w, h,
			fill=False,
			linestyle="-.",
			**kwargs
		))

	def __getitem__(self, index):
		if isinstance(index, slice):
			current, stop, step = index.indices(len(self))
			return [self.get_example(i) for i in
					six.moves.range(current, stop, step)]
		elif isinstance(index, list) or isinstance(index, np.ndarray):
			return [self.get_example(i) for i in index]
		else:
			return self.get_example(index)
